@@ -20,9 +20,9 @@
 
 #![recursion_limit = "256"]
 
-use quote::quote;
 use proc_macro::TokenStream;
-use syn::{parse_macro_input, DeriveInput, Data, DataStruct, Ident};
+use quote::quote;
+use syn::{parse_macro_input, Data, DataStruct, DeriveInput, Ident};
 
 /// Generates a delegating `NetworkBehaviour` implementation for the struct this is used for. See
 /// the trait documentation for better description.
@@ -45,42 +45,46 @@ fn build(ast: &DeriveInput) -> TokenStream {
 fn build_struct(ast: &DeriveInput, data_struct: &DataStruct) -> TokenStream {
     let name = &ast.ident;
     let (_, ty_generics, where_clause) = ast.generics.split_for_impl();
-    let multiaddr = quote!{::libp2p::core::Multiaddr};
-    let trait_to_impl = quote!{::libp2p::swarm::NetworkBehaviour};
-    let net_behv_event_proc = quote!{::libp2p::swarm::NetworkBehaviourEventProcess};
-    let either_ident = quote!{::libp2p::core::either::EitherOutput};
-    let network_behaviour_action = quote!{::libp2p::swarm::NetworkBehaviourAction};
-    let into_protocols_handler = quote!{::libp2p::swarm::IntoProtocolsHandler};
-    let protocols_handler = quote!{::libp2p::swarm::ProtocolsHandler};
-    let into_proto_select_ident = quote!{::libp2p::swarm::IntoProtocolsHandlerSelect};
-    let peer_id = quote!{::libp2p::core::PeerId};
-    let connection_id = quote!{::libp2p::core::connection::ConnectionId};
-    let connected_point = quote!{::libp2p::core::ConnectedPoint};
-    let listener_id = quote!{::libp2p::core::connection::ListenerId};
+    let multiaddr = quote! {::libp2p::core::Multiaddr};
+    let trait_to_impl = quote! {::libp2p::swarm::NetworkBehaviour};
+    let net_behv_event_proc = quote! {::libp2p::swarm::NetworkBehaviourEventProcess};
+    let either_ident = quote! {::libp2p::core::either::EitherOutput};
+    let network_behaviour_action = quote! {::libp2p::swarm::NetworkBehaviourAction};
+    let into_protocols_handler = quote! {::libp2p::swarm::IntoProtocolsHandler};
+    let protocols_handler = quote! {::libp2p::swarm::ProtocolsHandler};
+    let into_proto_select_ident = quote! {::libp2p::swarm::IntoProtocolsHandlerSelect};
+    let peer_id = quote! {::libp2p::core::PeerId};
+    let connection_id = quote! {::libp2p::core::connection::ConnectionId};
+    let dial_errors = quote! {Option<&Vec<::libp2p::core::Multiaddr>>};
+    let connected_point = quote! {::libp2p::core::ConnectedPoint};
+    let listener_id = quote! {::libp2p::core::connection::ListenerId};
+    let dial_error = quote! {::libp2p::swarm::DialError};
 
-    let poll_parameters = quote!{::libp2p::swarm::PollParameters};
+    let poll_parameters = quote! {::libp2p::swarm::PollParameters};
 
     // Build the generics.
     let impl_generics = {
         let tp = ast.generics.type_params();
         let lf = ast.generics.lifetimes();
         let cst = ast.generics.const_params();
-        quote!{<#(#lf,)* #(#tp,)* #(#cst,)*>}
+        quote! {<#(#lf,)* #(#tp,)* #(#cst,)*>}
     };
 
     // Whether or not we require the `NetworkBehaviourEventProcess` trait to be implemented.
     let event_process = {
-        let mut event_process = true; // Default to true for backwards compatibility
+        let mut event_process = false;
 
         for meta_items in ast.attrs.iter().filter_map(get_meta_items) {
             for meta_item in meta_items {
                 match meta_item {
-                    syn::NestedMeta::Meta(syn::Meta::NameValue(ref m)) if m.path.is_ident("event_process") => {
+                    syn::NestedMeta::Meta(syn::Meta::NameValue(ref m))
+                        if m.path.is_ident("event_process") =>
+                    {
                         if let syn::Lit::Bool(ref b) = m.lit {
                             event_process = b.value
                         }
                     }
-                    _ => ()
+                    _ => (),
                 }
             }
         }
@@ -92,17 +96,19 @@ fn build_struct(ast: &DeriveInput, data_struct: &DataStruct) -> TokenStream {
     // If we find a `#[behaviour(out_event = "Foo")]` attribute on the struct, we set `Foo` as
     // the out event. Otherwise we use `()`.
     let out_event = {
-        let mut out = quote!{()};
+        let mut out = quote! {()};
         for meta_items in ast.attrs.iter().filter_map(get_meta_items) {
             for meta_item in meta_items {
                 match meta_item {
-                    syn::NestedMeta::Meta(syn::Meta::NameValue(ref m)) if m.path.is_ident("out_event") => {
+                    syn::NestedMeta::Meta(syn::Meta::NameValue(ref m))
+                        if m.path.is_ident("out_event") =>
+                    {
                         if let syn::Lit::Str(ref s) = m.lit {
                             let ident: syn::Type = syn::parse_str(&s.value()).unwrap();
-                            out = quote!{#ident};
+                            out = quote! {#ident};
                         }
                     }
-                    _ => ()
+                    _ => (),
                 }
             }
         }
@@ -111,81 +117,95 @@ fn build_struct(ast: &DeriveInput, data_struct: &DataStruct) -> TokenStream {
 
     // Build the `where ...` clause of the trait implementation.
     let where_clause = {
-        let additional = data_struct.fields.iter()
+        let additional = data_struct
+            .fields
+            .iter()
             .filter(|x| !is_ignored(x))
             .flat_map(|field| {
                 let ty = &field.ty;
                 vec![
-                    quote!{#ty: #trait_to_impl},
+                    quote! {#ty: #trait_to_impl},
                     if event_process {
-                        quote!{Self: #net_behv_event_proc<<#ty as #trait_to_impl>::OutEvent>}
+                        quote! {Self: #net_behv_event_proc<<#ty as #trait_to_impl>::OutEvent>}
                     } else {
-                        quote!{#out_event: From< <#ty as #trait_to_impl>::OutEvent >}
-                    }
+                        quote! {#out_event: From< <#ty as #trait_to_impl>::OutEvent >}
+                    },
                 ]
             })
             .collect::<Vec<_>>();
 
         if let Some(where_clause) = where_clause {
             if where_clause.predicates.trailing_punct() {
-                Some(quote!{#where_clause #(#additional),*})
+                Some(quote! {#where_clause #(#additional),*})
             } else {
-                Some(quote!{#where_clause, #(#additional),*})
+                Some(quote! {#where_clause, #(#additional),*})
             }
         } else {
-            Some(quote!{where #(#additional),*})
+            Some(quote! {where #(#additional),*})
         }
     };
 
     // Build the list of statements to put in the body of `addresses_of_peer()`.
     let addresses_of_peer_stmts = {
-        data_struct.fields.iter().enumerate().filter_map(move |(field_n, field)| {
-            if is_ignored(&field) {
-                return None;
-            }
+        data_struct
+            .fields
+            .iter()
+            .enumerate()
+            .filter_map(move |(field_n, field)| {
+                if is_ignored(field) {
+                    return None;
+                }
 
-            Some(match field.ident {
-                Some(ref i) => quote!{ out.extend(self.#i.addresses_of_peer(peer_id)); },
-                None => quote!{ out.extend(self.#field_n.addresses_of_peer(peer_id)); },
+                Some(match field.ident {
+                    Some(ref i) => quote! { out.extend(self.#i.addresses_of_peer(peer_id)); },
+                    None => quote! { out.extend(self.#field_n.addresses_of_peer(peer_id)); },
+                })
             })
-        })
     };
 
     // Build the list of statements to put in the body of `inject_connected()`.
     let inject_connected_stmts = {
-        data_struct.fields.iter().enumerate().filter_map(move |(field_n, field)| {
-            if is_ignored(&field) {
-                return None;
-            }
-            Some(match field.ident {
-                Some(ref i) => quote!{ self.#i.inject_connected(peer_id); },
-                None => quote!{ self.#field_n.inject_connected(peer_id); },
+        data_struct
+            .fields
+            .iter()
+            .enumerate()
+            .filter_map(move |(field_n, field)| {
+                if is_ignored(field) {
+                    return None;
+                }
+                Some(match field.ident {
+                    Some(ref i) => quote! { self.#i.inject_connected(peer_id); },
+                    None => quote! { self.#field_n.inject_connected(peer_id); },
+                })
             })
-        })
     };
 
     // Build the list of statements to put in the body of `inject_disconnected()`.
     let inject_disconnected_stmts = {
-        data_struct.fields.iter().enumerate().filter_map(move |(field_n, field)| {
-            if is_ignored(&field) {
-                return None;
-            }
-            Some(match field.ident {
-                Some(ref i) => quote!{ self.#i.inject_disconnected(peer_id); },
-                None => quote!{ self.#field_n.inject_disconnected(peer_id); },
+        data_struct
+            .fields
+            .iter()
+            .enumerate()
+            .filter_map(move |(field_n, field)| {
+                if is_ignored(field) {
+                    return None;
+                }
+                Some(match field.ident {
+                    Some(ref i) => quote! { self.#i.inject_disconnected(peer_id); },
+                    None => quote! { self.#field_n.inject_disconnected(peer_id); },
+                })
             })
-        })
     };
 
     // Build the list of statements to put in the body of `inject_connection_established()`.
     let inject_connection_established_stmts = {
         data_struct.fields.iter().enumerate().filter_map(move |(field_n, field)| {
-            if is_ignored(&field) {
+            if is_ignored(field) {
                 return None;
             }
             Some(match field.ident {
-                Some(ref i) => quote!{ self.#i.inject_connection_established(peer_id, connection_id, endpoint); },
-                None => quote!{ self.#field_n.inject_connection_established(peer_id, connection_id, endpoint); },
+                Some(ref i) => quote!{ self.#i.inject_connection_established(peer_id, connection_id, endpoint, errors); },
+                None => quote!{ self.#field_n.inject_connection_established(peer_id, connection_id, endpoint, errors); },
             })
         })
     };
@@ -193,7 +213,7 @@ fn build_struct(ast: &DeriveInput, data_struct: &DataStruct) -> TokenStream {
     // Build the list of statements to put in the body of `inject_address_change()`.
     let inject_address_change_stmts = {
         data_struct.fields.iter().enumerate().filter_map(move |(field_n, field)| {
-            if is_ignored(&field) {
+            if is_ignored(field) {
                 return None;
             }
             Some(match field.ident {
@@ -205,146 +225,230 @@ fn build_struct(ast: &DeriveInput, data_struct: &DataStruct) -> TokenStream {
 
     // Build the list of statements to put in the body of `inject_connection_closed()`.
     let inject_connection_closed_stmts = {
-        data_struct.fields.iter().enumerate().filter_map(move |(field_n, field)| {
-            if is_ignored(&field) {
-                return None;
-            }
-            Some(match field.ident {
-                Some(ref i) => quote!{ self.#i.inject_connection_closed(peer_id, connection_id, endpoint); },
-                None => quote!{ self.#field_n.inject_connection_closed(peer_id, connection_id, endpoint); },
-            })
-        })
-    };
+        data_struct
+            .fields
+            .iter()
+            .enumerate()
+            // The outmost handler belongs to the last behaviour.
+            .rev()
+            .filter(|f| !is_ignored(f.1))
+            .enumerate()
+            .map(move |(enum_n, (field_n, field))| {
+                let handler = if field_n == 0 {
+                    // Given that the iterator is reversed, this is the innermost handler only.
+                    quote! { let handler = handlers }
+                } else {
+                    quote! {
+                        let (handlers, handler) = handlers.into_inner()
+                    }
+                };
+                let inject = match field.ident {
+                    Some(ref i) => quote!{ self.#i.inject_connection_closed(peer_id, connection_id, endpoint, handler) },
+                    None => quote!{ self.#enum_n.inject_connection_closed(peer_id, connection_id, endpoint, handler) },
+                };
 
-    // Build the list of statements to put in the body of `inject_addr_reach_failure()`.
-    let inject_addr_reach_failure_stmts = {
-        data_struct.fields.iter().enumerate().filter_map(move |(field_n, field)| {
-            if is_ignored(&field) {
-                return None;
-            }
-
-            Some(match field.ident {
-                Some(ref i) => quote!{ self.#i.inject_addr_reach_failure(peer_id, addr, error); },
-                None => quote!{ self.#field_n.inject_addr_reach_failure(peer_id, addr, error); },
+                quote! {
+                    #handler;
+                    #inject;
+                }
             })
-        })
     };
 
     // Build the list of statements to put in the body of `inject_dial_failure()`.
     let inject_dial_failure_stmts = {
-        data_struct.fields.iter().enumerate().filter_map(move |(field_n, field)| {
-            if is_ignored(&field) {
-                return None;
-            }
+        data_struct
+            .fields
+            .iter()
+            .enumerate()
+            // The outmost handler belongs to the last behaviour.
+            .rev()
+            .filter(|f| !is_ignored(f.1))
+            .enumerate()
+            .map(move |(enum_n, (field_n, field))| {
+                let handler = if field_n == 0 {
+                    // Given that the iterator is reversed, this is the innermost handler only.
+                    quote! { let handler = handlers }
+                } else {
+                    quote! {
+                        let (handlers, handler) = handlers.into_inner()
+                    }
+                };
 
-            Some(match field.ident {
-                Some(ref i) => quote!{ self.#i.inject_dial_failure(peer_id); },
-                None => quote!{ self.#field_n.inject_dial_failure(peer_id); },
+                let inject = match field.ident {
+                    Some(ref i) => {
+                        quote! { self.#i.inject_dial_failure(peer_id, handler, error.clone()) }
+                    }
+                    None => {
+                        quote! { self.#enum_n.inject_dial_failure(peer_id, handler, error.clone()) }
+                    }
+                };
+
+                quote! {
+                    #handler;
+                    #inject;
+                }
+            })
+    };
+
+    // Build the list of statements to put in the body of `inject_listen_failure()`.
+    let inject_listen_failure_stmts = {
+        data_struct
+            .fields
+            .iter()
+            .enumerate()
+            .rev()
+            .filter(|f| !is_ignored(f.1))
+            .enumerate()
+            .map(move |(enum_n, (field_n, field))| {
+                let handler = if field_n == 0 {
+                    quote! { let handler = handlers }
+                } else {
+                    quote! {
+                        let (handlers, handler) = handlers.into_inner()
+                    }
+                };
+
+                let inject = match field.ident {
+                    Some(ref i) => quote! { self.#i.inject_listen_failure(local_addr, send_back_addr, handler) },
+                    None => quote! { self.#enum_n.inject_listen_failure(local_addr, send_back_addr, handler) },
+                };
+
+                quote! {
+                    #handler;
+                    #inject;
+                }
             })
-        })
     };
 
     // Build the list of statements to put in the body of `inject_new_listener()`.
     let inject_new_listener_stmts = {
-        data_struct.fields.iter().enumerate().filter_map(move |(field_n, field)| {
-            if is_ignored(&field) {
-                return None;
-            }
+        data_struct
+            .fields
+            .iter()
+            .enumerate()
+            .filter_map(move |(field_n, field)| {
+                if is_ignored(field) {
+                    return None;
+                }
 
-            Some(match field.ident {
-                Some(ref i) => quote!{ self.#i.inject_new_listener(id); },
-                None => quote!{ self.#field_n.inject_new_listener(id); },
+                Some(match field.ident {
+                    Some(ref i) => quote! { self.#i.inject_new_listener(id); },
+                    None => quote! { self.#field_n.inject_new_listener(id); },
+                })
             })
-        })
     };
 
     // Build the list of statements to put in the body of `inject_new_listen_addr()`.
     let inject_new_listen_addr_stmts = {
-        data_struct.fields.iter().enumerate().filter_map(move |(field_n, field)| {
-            if is_ignored(&field) {
-                return None;
-            }
+        data_struct
+            .fields
+            .iter()
+            .enumerate()
+            .filter_map(move |(field_n, field)| {
+                if is_ignored(field) {
+                    return None;
+                }
 
-            Some(match field.ident {
-                Some(ref i) => quote!{ self.#i.inject_new_listen_addr(id, addr); },
-                None => quote!{ self.#field_n.inject_new_listen_addr(id, addr); },
+                Some(match field.ident {
+                    Some(ref i) => quote! { self.#i.inject_new_listen_addr(id, addr); },
+                    None => quote! { self.#field_n.inject_new_listen_addr(id, addr); },
+                })
             })
-        })
     };
 
     // Build the list of statements to put in the body of `inject_expired_listen_addr()`.
     let inject_expired_listen_addr_stmts = {
-        data_struct.fields.iter().enumerate().filter_map(move |(field_n, field)| {
-            if is_ignored(&field) {
-                return None;
-            }
+        data_struct
+            .fields
+            .iter()
+            .enumerate()
+            .filter_map(move |(field_n, field)| {
+                if is_ignored(field) {
+                    return None;
+                }
 
-            Some(match field.ident {
-                Some(ref i) => quote!{ self.#i.inject_expired_listen_addr(id, addr); },
-                None => quote!{ self.#field_n.inject_expired_listen_addr(id, addr); },
+                Some(match field.ident {
+                    Some(ref i) => quote! { self.#i.inject_expired_listen_addr(id, addr); },
+                    None => quote! { self.#field_n.inject_expired_listen_addr(id, addr); },
+                })
             })
-        })
     };
 
     // Build the list of statements to put in the body of `inject_new_external_addr()`.
     let inject_new_external_addr_stmts = {
-        data_struct.fields.iter().enumerate().filter_map(move |(field_n, field)| {
-            if is_ignored(&field) {
-                return None;
-            }
+        data_struct
+            .fields
+            .iter()
+            .enumerate()
+            .filter_map(move |(field_n, field)| {
+                if is_ignored(field) {
+                    return None;
+                }
 
-            Some(match field.ident {
-                Some(ref i) => quote!{ self.#i.inject_new_external_addr(addr); },
-                None => quote!{ self.#field_n.inject_new_external_addr(addr); },
+                Some(match field.ident {
+                    Some(ref i) => quote! { self.#i.inject_new_external_addr(addr); },
+                    None => quote! { self.#field_n.inject_new_external_addr(addr); },
+                })
             })
-        })
     };
 
     // Build the list of statements to put in the body of `inject_expired_external_addr()`.
     let inject_expired_external_addr_stmts = {
-        data_struct.fields.iter().enumerate().filter_map(move |(field_n, field)| {
-            if is_ignored(&field) {
-                return None;
-            }
+        data_struct
+            .fields
+            .iter()
+            .enumerate()
+            .filter_map(move |(field_n, field)| {
+                if is_ignored(field) {
+                    return None;
+                }
 
-            Some(match field.ident {
-                Some(ref i) => quote!{ self.#i.inject_expired_external_addr(addr); },
-                None => quote!{ self.#field_n.inject_expired_external_addr(addr); },
+                Some(match field.ident {
+                    Some(ref i) => quote! { self.#i.inject_expired_external_addr(addr); },
+                    None => quote! { self.#field_n.inject_expired_external_addr(addr); },
+                })
             })
-        })
     };
 
     // Build the list of statements to put in the body of `inject_listener_error()`.
     let inject_listener_error_stmts = {
-        data_struct.fields.iter().enumerate().filter_map(move |(field_n, field)| {
-            if is_ignored(&field) {
-                return None
-            }
-            Some(match field.ident {
-                Some(ref i) => quote!(self.#i.inject_listener_error(id, err);),
-                None => quote!(self.#field_n.inject_listener_error(id, err);)
+        data_struct
+            .fields
+            .iter()
+            .enumerate()
+            .filter_map(move |(field_n, field)| {
+                if is_ignored(field) {
+                    return None;
+                }
+                Some(match field.ident {
+                    Some(ref i) => quote!(self.#i.inject_listener_error(id, err);),
+                    None => quote!(self.#field_n.inject_listener_error(id, err);),
+                })
             })
-        })
     };
 
     // Build the list of statements to put in the body of `inject_listener_closed()`.
     let inject_listener_closed_stmts = {
-        data_struct.fields.iter().enumerate().filter_map(move |(field_n, field)| {
-            if is_ignored(&field) {
-                return None
-            }
-            Some(match field.ident {
-                Some(ref i) => quote!(self.#i.inject_listener_closed(id, reason);),
-                None => quote!(self.#field_n.inject_listener_closed(id, reason);)
+        data_struct
+            .fields
+            .iter()
+            .enumerate()
+            .filter_map(move |(field_n, field)| {
+                if is_ignored(field) {
+                    return None;
+                }
+                Some(match field.ident {
+                    Some(ref i) => quote!(self.#i.inject_listener_closed(id, reason);),
+                    None => quote!(self.#field_n.inject_listener_closed(id, reason);),
+                })
             })
-        })
     };
 
     // Build the list of variants to put in the body of `inject_event()`.
     //
     // The event type is a construction of nested `#either_ident`s of the events of the children.
     // We call `inject_event` on the corresponding child.
-    let inject_node_event_stmts = data_struct.fields.iter().enumerate().filter(|f| !is_ignored(&f.1)).enumerate().map(|(enum_n, (field_n, field))| {
+    let inject_node_event_stmts = data_struct.fields.iter().enumerate().filter(|f| !is_ignored(f.1)).enumerate().map(|(enum_n, (field_n, field))| {
         let mut elem = if enum_n != 0 {
             quote!{ #either_ident::Second(ev) }
         } else {
@@ -365,17 +469,18 @@ fn build_struct(ast: &DeriveInput, data_struct: &DataStruct) -> TokenStream {
     let protocols_handler_ty = {
         let mut ph_ty = None;
         for field in data_struct.fields.iter() {
-            if is_ignored(&field) {
+            if is_ignored(field) {
                 continue;
             }
             let ty = &field.ty;
-            let field_info = quote!{ <#ty as #trait_to_impl>::ProtocolsHandler };
+            let field_info = quote! { <#ty as #trait_to_impl>::ProtocolsHandler };
             match ph_ty {
-                Some(ev) => ph_ty = Some(quote!{ #into_proto_select_ident<#ev, #field_info> }),
+                Some(ev) => ph_ty = Some(quote! { #into_proto_select_ident<#ev, #field_info> }),
                 ref mut ev @ None => *ev = Some(field_info),
             }
         }
-        ph_ty.unwrap_or(quote!{()})     // TODO: `!` instead
+        // ph_ty = Some(quote! )
+        ph_ty.unwrap_or(quote! {()}) // TODO: `!` instead
     };
 
     // The content of `new_handler()`.
@@ -384,13 +489,13 @@ fn build_struct(ast: &DeriveInput, data_struct: &DataStruct) -> TokenStream {
         let mut out_handler = None;
 
         for (field_n, field) in data_struct.fields.iter().enumerate() {
-            if is_ignored(&field) {
+            if is_ignored(field) {
                 continue;
             }
 
             let field_name = match field.ident {
-                Some(ref i) => quote!{ self.#i },
-                None => quote!{ self.#field_n },
+                Some(ref i) => quote! { self.#i },
+                None => quote! { self.#field_n },
             };
 
             let builder = quote! {
@@ -398,29 +503,33 @@ fn build_struct(ast: &DeriveInput, data_struct: &DataStruct) -> TokenStream {
             };
 
             match out_handler {
-                Some(h) => out_handler = Some(quote!{ #into_protocols_handler::select(#h, #builder) }),
+                Some(h) => {
+                    out_handler = Some(quote! { #into_protocols_handler::select(#h, #builder) })
+                }
                 ref mut h @ None => *h = Some(builder),
             }
         }
 
-        out_handler.unwrap_or(quote!{()})     // TODO: incorrect
+        out_handler.unwrap_or(quote! {()}) // TODO: See test `empty`.
     };
 
     // The method to use to poll.
     // If we find a `#[behaviour(poll_method = "poll")]` attribute on the struct, we call
     // `self.poll()` at the end of the polling.
     let poll_method = {
-        let mut poll_method = quote!{std::task::Poll::Pending};
+        let mut poll_method = quote! {std::task::Poll::Pending};
         for meta_items in ast.attrs.iter().filter_map(get_meta_items) {
             for meta_item in meta_items {
                 match meta_item {
-                    syn::NestedMeta::Meta(syn::Meta::NameValue(ref m)) if m.path.is_ident("poll_method") => {
+                    syn::NestedMeta::Meta(syn::Meta::NameValue(ref m))
+                        if m.path.is_ident("poll_method") =>
+                    {
                         if let syn::Lit::Str(ref s) = m.lit {
                             let ident: Ident = syn::parse_str(&s.value()).unwrap();
-                            poll_method = quote!{#name::#ident(self, cx, poll_params)};
+                            poll_method = quote! {#name::#ident(self, cx, poll_params)};
                         }
                     }
-                    _ => ()
+                    _ => (),
                 }
             }
         }
@@ -430,7 +539,7 @@ fn build_struct(ast: &DeriveInput, data_struct: &DataStruct) -> TokenStream {
     // List of statements to put in `poll()`.
     //
     // We poll each child one by one and wrap around the output.
-    let poll_stmts = data_struct.fields.iter().enumerate().filter(|f| !is_ignored(&f.1)).enumerate().map(|(enum_n, (field_n, field))| {
+    let poll_stmts = data_struct.fields.iter().enumerate().filter(|f| !is_ignored(f.1)).enumerate().map(|(enum_n, (field_n, field))| {
         let field_name = match field.ident {
             Some(ref i) => quote!{ self.#i },
             None => quote!{ self.#field_n },
@@ -445,6 +554,42 @@ fn build_struct(ast: &DeriveInput, data_struct: &DataStruct) -> TokenStream {
             wrapped_event = quote!{ #either_ident::First(#wrapped_event) };
         }
 
+        // `Dial` provides a handler of the specific behaviour triggering the
+        // event. Though in order for the final handler to be able to handle
+        // protocols of all behaviours, the provided handler needs to be
+        // combined with handlers of all other behaviours.
+        let provided_handler_and_new_handlers = {
+            let mut out_handler = None;
+
+            for (f_n, f) in data_struct.fields.iter().enumerate() {
+                if is_ignored(f) {
+                    continue;
+                }
+
+                let f_name = match f.ident {
+                    Some(ref i) => quote! { self.#i },
+                    None => quote! { self.#f_n },
+                };
+
+                let builder = if field_n == f_n {
+                    // The behaviour that triggered the event. Thus, instead of
+                    // creating a new handler, use the provided handler.
+                    quote! { provided_handler }
+                } else {
+                    quote! { #f_name.new_handler() }
+                };
+
+                match out_handler {
+                    Some(h) => {
+                        out_handler = Some(quote! { #into_protocols_handler::select(#h, #builder) })
+                    }
+                    ref mut h @ None => *h = Some(builder),
+                }
+            }
+
+            out_handler.unwrap_or(quote! {()}) // TODO: See test `empty`.
+        };
+
         let generate_event_match_arm = if event_process {
             quote! {
                 std::task::Poll::Ready(#network_behaviour_action::GenerateEvent(event)) => {
@@ -463,11 +608,8 @@ fn build_struct(ast: &DeriveInput, data_struct: &DataStruct) -> TokenStream {
             loop {
                 match #trait_to_impl::poll(&mut #field_name, cx, poll_params) {
                     #generate_event_match_arm
-                    std::task::Poll::Ready(#network_behaviour_action::DialAddress { address }) => {
-                        return std::task::Poll::Ready(#network_behaviour_action::DialAddress { address });
-                    }
-                    std::task::Poll::Ready(#network_behaviour_action::DialPeer { peer_id, condition }) => {
-                        return std::task::Poll::Ready(#network_behaviour_action::DialPeer { peer_id, condition });
+                    std::task::Poll::Ready(#network_behaviour_action::Dial { opts, handler: provided_handler }) => {
+                        return std::task::Poll::Ready(#network_behaviour_action::Dial { opts, handler: #provided_handler_and_new_handlers });
                     }
                     std::task::Poll::Ready(#network_behaviour_action::NotifyHandler { peer_id, handler, event }) => {
                         return std::task::Poll::Ready(#network_behaviour_action::NotifyHandler {
@@ -489,7 +631,7 @@ fn build_struct(ast: &DeriveInput, data_struct: &DataStruct) -> TokenStream {
     });
 
     // Now the magic happens.
-    let final_quote = quote!{
+    let final_quote = quote! {
         impl #impl_generics #trait_to_impl for #name #ty_generics
         #where_clause
         {
@@ -515,7 +657,7 @@ fn build_struct(ast: &DeriveInput, data_struct: &DataStruct) -> TokenStream {
                 #(#inject_disconnected_stmts);*
             }
 
-            fn inject_connection_established(&mut self, peer_id: &#peer_id, connection_id: &#connection_id, endpoint: &#connected_point) {
+            fn inject_connection_established(&mut self, peer_id: &#peer_id, connection_id: &#connection_id, endpoint: &#connected_point, errors: #dial_errors) {
                 #(#inject_connection_established_stmts);*
             }
 
@@ -523,16 +665,16 @@ fn build_struct(ast: &DeriveInput, data_struct: &DataStruct) -> TokenStream {
                 #(#inject_address_change_stmts);*
             }
 
-            fn inject_connection_closed(&mut self, peer_id: &#peer_id, connection_id: &#connection_id, endpoint: &#connected_point) {
+            fn inject_connection_closed(&mut self, peer_id: &#peer_id, connection_id: &#connection_id, endpoint: &#connected_point, handlers: <Self::ProtocolsHandler as #into_protocols_handler>::Handler) {
                 #(#inject_connection_closed_stmts);*
             }
 
-            fn inject_addr_reach_failure(&mut self, peer_id: Option<&#peer_id>, addr: &#multiaddr, error: &dyn std::error::Error) {
-                #(#inject_addr_reach_failure_stmts);*
+            fn inject_dial_failure(&mut self, peer_id: Option<#peer_id>, handlers: Self::ProtocolsHandler, error: &#dial_error) {
+                #(#inject_dial_failure_stmts);*
             }
 
-            fn inject_dial_failure(&mut self, peer_id: &#peer_id) {
-                #(#inject_dial_failure_stmts);*
+            fn inject_listen_failure(&mut self, local_addr: &#multiaddr, send_back_addr: &#multiaddr, handlers: Self::ProtocolsHandler) {
+                #(#inject_listen_failure_stmts);*
             }
 
             fn inject_new_listener(&mut self, id: #listener_id) {
@@ -574,10 +716,10 @@ fn build_struct(ast: &DeriveInput, data_struct: &DataStruct) -> TokenStream {
                 }
             }
 
-            fn poll(&mut self, cx: &mut std::task::Context, poll_params: &mut impl #poll_parameters) -> std::task::Poll<#network_behaviour_action<<<Self::ProtocolsHandler as #into_protocols_handler>::Handler as #protocols_handler>::InEvent, Self::OutEvent>> {
+            fn poll(&mut self, cx: &mut std::task::Context, poll_params: &mut impl #poll_parameters) -> std::task::Poll<#network_behaviour_action<Self::OutEvent, Self::ProtocolsHandler>> {
                 use libp2p::futures::prelude::*;
                 #(#poll_stmts)*
-                let f: std::task::Poll<#network_behaviour_action<<<Self::ProtocolsHandler as #into_protocols_handler>::Handler as #protocols_handler>::InEvent, Self::OutEvent>> = #poll_method;
+                let f: std::task::Poll<#network_behaviour_action<Self::OutEvent, Self::ProtocolsHandler>> = #poll_method;
                 f
             }
         }
@@ -609,7 +751,7 @@ fn is_ignored(field: &syn::Field) -> bool {
                 syn::NestedMeta::Meta(syn::Meta::Path(ref m)) if m.is_ident("ignore") => {
                     return true;
                 }
-                _ => ()
+                _ => (),
             }
         }
     }
